use std::{collections::HashMap, path::{Path, PathBuf}};

use crate::ipc::Message;

use self::plugin::Plugin;


mod plugin;

pub struct PlugMan {
    plugins: Vec<Plugin>,
    len_cap: usize,
}


impl PlugMan {
    pub fn new(len_cap: usize) -> Self {
        Self {
            plugins: Vec::new(),
            len_cap
        }
    }

    pub fn load<P: AsRef<Path>>(&mut self, dir: P) -> Result<(), PluginError>{
        let dir: &Path = dir.as_ref();
        log::debug!("Loading plugins from {:?}", dir);

        let files = std::fs::read_dir(dir)?;

        for file in files {
            let entry = file?;
            if entry.file_type()?.is_file() {
                let plugin = Plugin::load(entry.path().to_path_buf())?;
                log::info!("Loaded plugin '{}' ({}) licensed with {}", 
                    plugin.name().clone(),
                    plugin.version().clone().unwrap_or("None".to_string()),
                    plugin.license().clone().unwrap_or("None".to_string())
                );
                self.plugins.push(plugin);
            } else 
            if entry.file_type()?.is_dir() {
                self.load(entry.path().to_path_buf())?;
            }
        }
        if self.plugins.len() == 0 {
            log::error!("No plugins found");
            return Err(PluginError::NoPlugins);
        }

        log::info!("Loaded {} plugins", self.plugins.len());

        Ok(())
    }

    pub fn init_plugins(&mut self, cfg: &crate::config::Config) {
        for plugin in &mut self.plugins {
            plugin.init(cfg);
        }
    }
    
    #[allow(dead_code)]
    pub fn reload_plugins(&mut self) {
        for plugin in &mut self.plugins {
            if let Err(e) = plugin.reload() {
                log::error!("Failed to reload plugin {:?}: {e}", plugin.name()); 
            }
        }
    }

    #[allow(dead_code)]
    pub fn unload_plugins(&mut self) {
        while let Some(plug) = self.plugins.pop() {
            plug.free();
        }
    }

    pub fn poll_plugins(&mut self) -> HashMap<String, PolledText> {
        let mut answers = HashMap::new();
        for plugin in &mut self.plugins {
            if !plugin.enabled() {
                continue;
            }
            match plugin.poll() {
                Ok(v) => {
                    let pth = (plugin.name().clone(), PolledText::new(v));
                    if !pth.1.text().is_empty() {
                        answers.insert(pth.0, pth.1);
                    }
                },
                Err(e) => log::error!("Failed to poll plugin: {e}"),
            }
            
        }
        answers
    }

    pub fn send_msg(&mut self, msg: Message) {
        for plugin in &mut self.plugins {
            if let Some(pn) = &msg.plugin {
                if *pn != plugin.name() {
                    continue;
                }
            }
            plugin.send_msg(msg);
            break;
        }
    }
}

#[derive(Debug, Clone)]
pub struct PolledText {
    text: String,
    disabled: bool
}

impl PolledText {
    pub fn new(text: String) -> Self {
        Self { text, disabled: false }
    }
    pub fn text(&self) -> &String {
        &self.text
    }
    pub fn disabled(&self) -> bool {
        self.disabled
    } 
    pub fn disable(&mut self) {
        self.disabled = true;
    }
}

#[derive(Debug)]
pub enum PluginError {
    DlOpenError,
    IoError,
    NoPlugins
}

impl From<dlopen::Error> for PluginError {
    fn from(_: dlopen::Error) -> Self {
        Self::DlOpenError
    }
}

impl From<std::io::Error> for PluginError {
    fn from(_: std::io::Error) -> Self {
        Self::IoError
    }
}
