use std::{alloc::Layout, ffi::{c_char, CStr}, path::PathBuf};

use dlopen::raw::Library;

use crate::{config::Config, ipc::{Message, MessageFFI}};

use self::context::{PluginContext, PluginContextContainer};

pub mod loader;
pub mod context;

const MEM_LAYOUT: Result<Layout, std::alloc::LayoutError> = Layout::from_size_align(2048, 1);

#[repr(C)]
pub struct PluginInfo {
    name: *const c_char,
    version: *const c_char,
    license: *const c_char,
}

#[derive(Debug)]
pub struct Plugin {
    syms: Option<loader::PluginSyms>,
    path: PathBuf,
    lib: Option<Library>,
    enabled: bool,
    buf: *mut u8
}



impl Plugin {
    pub fn load(path: PathBuf) -> Result<Self, dlopen::Error> {
        let buf = unsafe {
            std::alloc::alloc_zeroed(MEM_LAYOUT.unwrap())
        };

        let mut s = Self {
            enabled: true,
            path,
            syms: None,
            lib: None,
            buf
        };
        s.reload_symbols()?;
        Ok(s)
    }
  
    pub fn syms(&self) -> loader::PluginSyms {
        self.syms.unwrap()
    }

    pub fn init(&mut self, cfg: &Config) {
        let conf_dir = &cfg.config_dir.join(self.name()).to_string_lossy().to_string();
        let _ = std::fs::create_dir(&conf_dir); // dont care 
        let ctx = PluginContextContainer::new(self, conf_dir);
        
        unsafe {
            if (self.syms().init)(&ctx.inner as *const PluginContext) != 0 {
                log::warn!("`init` had an error, disabling");
                self.disable();
            }
        }
    }
    pub fn reload(&mut self) -> Result<(), dlopen::Error> { 
        unsafe {
            let state = (self.syms().pre_reload)();
            
            if state.is_null() {
                log::warn!("`pre_reload` had an error, disabling");
                self.disable();
            }

            if let Err(e) = self.reload_symbols() {
                self.disable();
                return Err(e);
            }
            if (self.syms().post_reload)(state) != 0 {
                log::warn!("`post_reload` had an error, disabling");
                self.disable();
            }
        
        }
        Ok(())
    }

    pub fn send_msg(&mut self, msg: Message) {
        unsafe {
            // log::debug!("Sending message to '{}': {:?}", self.name(), msg);
            if (self.syms().on_msg)(&msg.ffi  as *const MessageFFI) != 0 {
                log::warn!("`on_msg` had an error, disabling");
                self.disable();
            }
        }
    }

    pub fn poll(&mut self) -> Result<String, std::string::FromUtf8Error> {
        self.send_msg(Message::from_string("poll"));
        let s = unsafe {
            CStr::from_ptr(self.buf as *const c_char).to_string_lossy().to_string()
        }; 
        
        if !s.is_empty() {
            // log::debug!("Polled: {}", s);
        }

        Ok(s)
    }


    pub fn free(mut self) {
        unsafe {
            if (self.syms().free)() != 0 {
                log::warn!("`free` had an error, disabling");
                self.disable(); // Literaly useless but eh
            }
            std::alloc::dealloc(self.buf, MEM_LAYOUT.unwrap());
        }
        drop(self.lib.unwrap()); 
    }

    pub fn enabled(&self) -> bool {
        self.enabled
    }

    pub fn disable(&mut self) {
        self.enabled = false;
    }

    pub fn name(&self) -> String {
        unsafe {
            let info = (self.syms().get_info)();
            CStr::from_ptr((*info).name).to_string_lossy().to_string()
        }
    }

    pub fn version(&self) -> Option<String> {
        unsafe {
            let info = (self.syms().get_info)();
            let v = (*info).version;
            if v.is_null() {
                return None;
            }
            Some(CStr::from_ptr(v).to_string_lossy().to_string())
        }
    }

    pub fn license(&self) -> Option<String> {
        unsafe {
            let info = (self.syms().get_info)();
            let v = (*info).license;
            if v.is_null() {
                return None;
            }
            Some(CStr::from_ptr(v).to_string_lossy().to_string())
        }
    }
}

